/*
 * Jednoducha implementace ovladace pevneho disku pracujici
 * v rezimu PIO s adresaci LBA28.
 */

#include <stdint.h>

#include "ata.h"

#include "../util/io.h"
#include "../util/util.h"

#define ATA_SOFT_RESET_ATTEMPTS	(1000)

#define	DATA_REG          (0)
#define STATUS_REG        (7)

static int ata_soft_reset(struct dev_ata *dev);

/**
 * Inicializuje zarizeni na primarni sbernici
 */
int ata_init(struct dev_ata *dev, int bus, int slave)
{
	dev->io_base  = (bus == ATA_PRIMARY_BUS) ? 0x1f0 : 0x170;
	dev->ctl_base = (bus == ATA_PRIMARY_BUS) ? 0x3f6 : 0x376;
	dev->slave    = slave;
	ata_soft_reset(dev);
	return 0;
}

/**
 * Do ciloveho bufferu nacte sector_count sektoru.
 * Argument lba - udava adresu sektoru, odkud se bude cist.
 * Predpoklada se velikost sektoru 512 B.
 */
int ata_read(struct dev_ata *dev, uint32_t lba, uint8_t sector_count, unsigned char *buf)
{
	uint8_t status;
	uint8_t slave = dev->slave ? 1 << 4 : 0;
	outb(dev->io_base + 2, sector_count);
	outb(dev->io_base + 3, lba & 0xff);         /* LBAlo */
	outb(dev->io_base + 4, (lba >> 8) & 0xff);  /* LBAmid */
	outb(dev->io_base + 5, (lba >> 16) & 0xff); /* LBAhi */
	outb(dev->io_base + 6, ((lba >> 24) & 0x0f) | 0xe0 | slave); /* LBAhi + 0xe0 nastaveni modu LBA + bit pro slave */
	outb(dev->io_base + 7, 0x20); /* read command */

	/* prvni pokusy o cteni stavu nemusi projit (zpozdeni 400ns) */
	for (int i = 0; i < 4; i++) {
		status = inb(dev->io_base + STATUS_REG);
		if (status & 0x80) continue;   /* BSY */
		if (status & 0x08) break;      /* DRQ */
	}

	for (int sector = 0; sector < sector_count; sector++) {
		/* pred kazdym ctenim sektoru je potreba testovat dostupnost dat */
		do {
			status = inb(dev->io_base + STATUS_REG);
		} while (status & 0x80); /* BSY */
		if (status & 0x21) return EATA_DEV_FAIL; /* ERR nebo DF */		

		/* prenos dat */
		for (int i = 0; i < 256; i++) {
			uint16_t data = inw(dev->io_base + DATA_REG);
			*(buf++) = (data & 0xff);
			*(buf++) = (data >> 8) & 0xff;
		}
	}
	return EOK;
}

/**
 * Na adresu danou argumentem lba ulozi obsah buf.
 * Velikost zapisovanych dat je dana jako sector_count * 512 B.
 */

int ata_write(struct dev_ata *dev, uint32_t lba, uint8_t sector_count, unsigned char *buf)
{
	uint8_t status;
	uint8_t slave = dev->slave ? 1 << 4 : 0;
	outb(dev->io_base + 2, sector_count);
	outb(dev->io_base + 3, lba & 0xff);         /* LBAlo */
	outb(dev->io_base + 4, (lba >> 8) & 0xff);  /* LBAmid */
	outb(dev->io_base + 5, (lba >> 16) & 0xff); /* LBAhi */
	outb(dev->io_base + 6, ((lba >> 24) & 0x0f) | 0xe0 | slave); /* LBAhi + 0xe0 nastaveni modu LBA + bit pro slave */
	outb(dev->io_base + 7, 0x30); /* write command */

	/* prvni pokusy o cteni stavu nemusi projit (zpozdeni 400ns) */
	for (int i = 0; i < 4; i++) {
		status = inb(dev->io_base + STATUS_REG);
		if (status & 0x80) continue;   /* BSY */
	}

	for (int sector = 0; sector < sector_count; sector++) {
		/* pred kazdym zapisem sektoru je potreba testovat dostupnost */
		do {
			status = inb(dev->io_base + STATUS_REG);
		} while (status & 0x80); /* BSY */
		if (status & 0x21) return EATA_DEV_FAIL; /* ERR nebo DF */		

		/* prenos dat */
		for (int i = 0; i < 256; i++) {
			uint16_t data = *buf | (*(buf + 1) << 8);
			outw(dev->io_base + DATA_REG, data);
			buf += 2;
		}
	}
	return EOK;
}


/**
 * Vraci, jaky typ zarizeni je pripojeny na primarni sbernici
 */
int ata_detect(int bus, int slave)
{
	struct dev_ata dev;
	ata_init(&dev, bus, slave);

	outb(dev.io_base + 6, 0xa0 | slave << 4);

	/* zpozdeni 400 ns */
	inb(dev.ctl_base);
	inb(dev.ctl_base);
	inb(dev.ctl_base);
	inb(dev.ctl_base);

	/* ziska identifikaci (signaturu) zarizeni */
	unsigned sigl = inb(dev.io_base + 4);
	unsigned sigh = inb(dev.io_base + 5);

	if ((sigl == 0x14) && (sigh == 0xeb)) return ATA_PATAPI;
	if ((sigl == 0x69) && (sigh == 0x96)) return ATA_SATAPI;
	if ((sigl == 0x00) && (sigh == 0x00)) return ATA_PATA;
	if ((sigl == 0x3c) && (sigh == 0xc3)) return ATA_SATA;
	return ATA_UNKNOWN;
}


/**
 * Vraci velikost zarizeni v blocich
 */
int ata_drive_size(struct dev_ata *dev)
{
	uint8_t status;
	uint8_t slave = dev->slave ? 1 << 4 : 0;
	uint16_t buf[256];
	outb(dev->io_base + 2, 0);
	outb(dev->io_base + 3, 0);  /* LBAlo */
	outb(dev->io_base + 4, 0);  /* LBAmid */
	outb(dev->io_base + 5, 0);  /* LBAhi */
	outb(dev->io_base + 6, 0xa0 | slave); /* bit pro slave */
	outb(dev->io_base + 7, 0xec); /* identify command */

	/* prvni pokusy o cteni stavu nemusi projit (zpozdeni 400ns) */
	for (int i = 0; i < 4; i++) {
		status = inb(dev->io_base + STATUS_REG);
		if (status & 0x80) continue;   /* BSY */
		if (status & 0x08) break;      /* DRQ */
	}

	/* pred kazdym ctenim je potreba testovat dostupnost dat */
	do {
		status = inb(dev->io_base + STATUS_REG);
	} while (status & 0x80); /* BSY */
	if (status & 0x21) return EATA_DEV_FAIL; /* ERR nebo DF */		

	/* prenos dat */
	for (int i = 0; i < 256; i++) {
		buf[i] = inw(dev->io_base + DATA_REG);
	}
	return ((int) buf[60]) | (((int) buf[61]) << 16);
}

/**
 * Provede reset zarizeni na primarni sbernici, nutne pred zacatkem komunikace
 */
static int ata_soft_reset(struct dev_ata *dev)
{
	outb(dev->ctl_base, 0x06); /* nastav bit pro software reset, nIEN -- blokuj preruseni */
	outb(dev->ctl_base, 0x02); /* vynuluj bit pro software reset, nIEN -- blokuj prerueseni */
	
	/* nekolikrat naprazdno zkusi precist stav
	 * specifikace vyzaduje zpozdeni 400ns */
	inb(dev->ctl_base); 
	inb(dev->ctl_base); 
	inb(dev->ctl_base); 
	inb(dev->ctl_base);

	for (int i = 0; i < ATA_SOFT_RESET_ATTEMPTS; i++) {
		uint8_t status = inb(dev->ctl_base);	
		status &= 0xc0;       /* priznaky BSY a RDY */
		if (status == 0x40)   /* BSY = 0, RDY = 1   */
			return EOK;
	}
	return EATA_DEV_FAIL;
}
