/*
 *
 * Sada funkci pro obsluhu klavesnice
 *
 */

#include <stdint.h>
#include "keyb.h"

#include "../util/io.h"

#define KEYMAP_SIZE (0x80 * 2)
#define KEY_SHIFT_LEFT	(0x2a)
#define KEY_SHIFT_RIGHT	(0x36)
#define KEY_CTRL	(0x1d)
#define KEY_RELEASED	(0x80)

/* aktivni modifikatory, viz MOD_EXT, MOD_SHIFT, ... */
unsigned int keyb_mode = 0;

uint16_t keymap[KEYMAP_SIZE];

uint16_t keymap_us[] = { 
	'\0',      '\0',     /* unused */
	E(ESCAPE), S(E(ESCAPE)),
	'1',       '!',		 
	'2',       '@',
	'3',       '#',
	'4',       '$',
	'5',       '%',
	'6',       '^',
	'7',       '&',
	'8',       '*',
	'9',       '(',
	'0',       ')',
	'-',       '_',
	'=',       '+',
	'\b',      S('\b'),
	'\t',      S('\t'),
	'Q',       S('Q'),
	'W',       S('W'),
	'E',       S('E'),
	'R',       S('R'),
	'T',       S('T'),
	'Y',       S('Y'),
	'U',       S('U'),
	'I',       S('I'),
	'O',       S('O'),
	'P',       S('P'),
	'[',       '{',
	']',       '}',
	'\n',      S('\n'),
	'\0',      '\0',    /* left control */
	'A',       S('A'),
	'S',       S('S'),
	'D',       S('D'),
	'F',       S('F'),
	'G',       S('G'),
	'H',       S('H'),
	'J',       S('J'),
	'K',       S('K'),
	'L',       S('L'),
	';',       ':', 
	'\'',      '"',
	'`',       '~',
	'\0',      '\0',   /* left shift*/
	'\\',      '|', 
	'Z',       S('Z'), 
	'X',       S('X'), 
	'C',       S('C'), 
	'V',       S('V'), 
	'B',       S('B'), 
	'N',       S('N'), 
	'M',       S('M'), 
	',',       '<',
	'.',       '>',
	'/',       '?',
	'\0',      '\0',   /* right shift */
	'*',       S('*'), /* keypad */
	'\0',      '\0',   /* left alt */
	' ',       S(' '),
	'\0',      '\0',   /* CapsLock */ 
	E(F1),      S(E(F1)), 
	E(F2),      S(E(F2)), 
	E(F3),      S(E(F3)), 
	E(F4),      S(E(F4)), 
	E(F5),      S(E(F5)), 
	E(F6),      S(E(F6)), 
	E(F7),      S(E(F7)), 
	E(F8),      S(E(F8)), 
	E(F9),      S(E(F9)), 
	E(F10),     S(E(F10)), 
	'\0',       '\0',   /* NumLock */
	'\0',       '\0',   /* ScrollLock */
	'7',        S('7'), /* keypad */
	'8',        S('8'), /* keypad */
	'9',        S('9'), /* keypad */
	'-',        S('-'), /* keypad */
	'4',        S('4'), /* keypad */
	'5',        S('5'), /* keypad */
	'6',        S('6'), /* keypad */
	'+',        S('+'), /* keypad */
	'1',        S('1'), /* keypad */
	'2',        S('2'), /* keypad */
	'3',        S('3'), /* keypad */
	'0',        S('0'), /* keypad */
	'.',        S('.')  /* keypad */
};



void keyb_init()
{
	for (int i = 0; i < KEYMAP_SIZE; i++)
		keymap[i] = 0;

	for (int i = 0; i < sizeof(keymap_us) / 2; i++)
		keymap[i] = keymap_us[i];
}

/**
 * Vraci scan code prave stisknute nebo uvolnene klavesy.
 * Blokujici operace.
 */
unsigned char keyb_scancode()
{
	while (1) {
		uint8_t status = inb(0x64);
		if (status & 0x01) {
			uint8_t scan_code = inb(0x60);
			return scan_code;
		}
		task_kernel_yield();
	}
}


/**
 * Prevede dany scancode na kombinaci stisknutych klaves,
 * bere v uvahu chovani rozsirujicich klaves
 */
static uint16_t keymap_get(char scan_code)
{
	uint16_t key_stroke = 0;
	if (keyb_mode & MOD_SHIFT) key_stroke = keymap[scan_code * 2 + 1];
	else key_stroke = keymap[scan_code * 2];
	if (key_stroke) {
		if (keyb_mode & MOD_CTRL) key_stroke = C(key_stroke);
		return key_stroke;
	}
	return 0;
}

/**
 * Funkce vracejici stisknute klavesy + modifikatory
 * Spodnich osm bitu je znak, zbyvajici horni bity obsahuji
 * priznaky jednotlivych modifikatoru MOD_*
 */
uint16_t keyb_get_stroke()
{
	while (1) {
		unsigned char scan_code = keyb_scancode();
		if (!(scan_code & KEY_RELEASED)) {
			switch (scan_code) {
				case KEY_SHIFT_LEFT:
				case KEY_SHIFT_RIGHT:
					keyb_mode |= MOD_SHIFT;
					break;
				case KEY_CTRL:
					keyb_mode |= MOD_CTRL;
					break;
				default: {
					uint16_t key_stroke = keymap_get(scan_code);
					if (key_stroke) return key_stroke;
				}
			}
		} else {
			switch (scan_code) {
				case (KEY_RELEASED | KEY_SHIFT_LEFT):
				case (KEY_RELEASED | KEY_SHIFT_RIGHT):
					keyb_mode &= ~MOD_SHIFT;
					break;
				case (KEY_RELEASED | KEY_CTRL):
					keyb_mode &= ~MOD_CTRL;
					break;
			}
		}
	}
}

/**
 * Funkce vracejici jednotlive stisknute znaky
 */
char keyb_getc()
{
	char c = '\0';
	do {
		uint16_t key_stroke = keyb_get_stroke();
		if (!IS_MOD_CTRL(key_stroke) && !IS_MOD_EXT(key_stroke)) {
			// preskocime rozsirujici znaky
			c = key_stroke & 0xff;
			if ((c >= 'A') && (c <= 'Z') && !IS_MOD_SHIFT(key_stroke)) {
				c += ('a' - 'A');
			}
		}
	} while (!c);
	return c;
}
