/*
 * Naprosto minimalisticka implementace podmnoziny libc
 * pro demonstraci moznosti operacnich systemu.
 */
#include <stdarg.h>

#include "../include/stdio.h"
#include "../include/stdlib.h"
#include "../include/unistd.h"
#include "../include/string.h"

#include "syscalls.h"

#define MAX_PATH_LEN	(255)

struct _stdio_file _stdout = { 1, 0, 0, 0, 0};
struct _stdio_file _stdin  = { 0, 0, 0, 0, 0 };

/*
 *
 * Prace se soubory
 *
 */

int open(const char *path, int flags, unsigned int mode)
{
	return sys_open(path, flags, mode);
}

int close(int fd)
{
	return sys_close(fd);
}

FILE *fopen(const char *path, const char *mode)
{
	/* TODO: nepodporuje kombinovany mod, cteni/zapis */
	if (strchr(mode, '+')) return NULL;

	/* TODO: podpora zapisu za konec souboru */
	if (strchr(mode, 'a')) return NULL;


	/* pokud existuje soubor, je zkracen na nulovou velikost */
	if (strchr(mode, 'w')) {
		sys_truncate(path, 0);
	}

	int fd = open(path, 0, 0);
	if (!fd) return NULL;

	FILE *f = malloc(sizeof(FILE));
	f->fd          = fd;	
	f->outbuf_size = 0;
        f->inbuf_size  = 0;
        f->inbuf_idx   = 0;
        f->eof         = 0;

	return f;	
}

int fclose(FILE *stream)
{
	int status = close(stream->fd);
	free(stream);
	if (status != 0) return EOF;
	return 0;
}

/*
 *
 * Prace s vystupy
 *
 */

static inline int _stdio_flush(struct _stdio_file *f)
{
	int status = sys_write(f->fd, f->outbuf, f->outbuf_size);
	f->outbuf_size = 0;
	return status;
}

static inline int _stdio_putc(struct _stdio_file *f, char c)
{
	f->outbuf[f->outbuf_size++] = c;
	if (f->outbuf_size >= _STDIO_OUTBUF_SIZE) {
		int status = _stdio_flush(f);
		if (status < 0) return EOF;
	}
	return c;
}

static inline int _stdio_puts(struct _stdio_file *f, char *s)
{
	while (*s) {
		int status = _stdio_putc(f, *s);
		if (status < 0) return EOF;
		s++;
	}
	return 0;
}
/**
 * Prevede cislo n na jeho textovou reprezentaci pri zakladu base
 * a vysledek ulozi do bufferu
 */
static void uint_to_str(char *buf, unsigned int n, unsigned char base)
{
        static char *map = "0123456789abcdef";
        char revbuf[33];
        unsigned int size = 0;
        do {
                revbuf[size] = map[n % base];
                n /= base;
                size++;
        } while (n);

        for (int i = 0; i < size; i++) {
                buf[i] = revbuf[size - i - 1];
        }
        buf[size] = '\0';
}

static int _stdio_puti(struct _stdio_file *f, int n, unsigned char base, int signed_int)
{
	char buf[33];

	int sign = signed_int && n < 0;
	if (signed_int && sign) n = -n;

	if (sign) {
		int status = _stdio_putc(f, '-');
		if (status < 0) return EOF;
	}
	uint_to_str(buf, n, base);
	return _stdio_puts(f, buf);

}

static int _stdio_printf(struct _stdio_file *f, const char *fmt, va_list ap)
{
	while (*fmt) {
		int status;
		if (*fmt == '%') {
			fmt++;;
			switch (*fmt) {
			case '%': status = _stdio_putc(f, '%'); break;
			case 'c': status = _stdio_putc(f, va_arg(ap, int)); break;
			case 's': status = _stdio_puts(f, va_arg(ap, char *)); break;
			case 'i': status = _stdio_puti(f, va_arg(ap, int), 10, 1); break;
			case 'u': status = _stdio_puti(f, va_arg(ap, int), 10, 0); break;
			case 'x': status = _stdio_puti(f, va_arg(ap, int), 16, 0); break;
			default: va_arg(ap, int); status = _stdio_putc(f, '?');
			}
		} else {	
			status = _stdio_putc(f, *fmt);
		}
		if (status < 0) return EOF;
		fmt++;
	}
	return (_stdio_flush(&_stdout) > 0) ? 0 : EOF;
}


int printf(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	int status = _stdio_printf(&_stdout, fmt, ap);
	va_end(ap);
	return status;
}

int fprintf(FILE *stream, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	int status = _stdio_printf(stream, fmt, ap);
	va_end(ap);
	return status;
}

int puts(char *s)
{
	if (_stdio_puts(&_stdout, s) < 0) return EOF;
	int status = _stdio_putc(&_stdout, '\n');
	if (status < 0) return EOF;
	return (_stdio_flush(&_stdout) > 0) ? 0 : EOF;
}

int putchar(int c)
{
	int status = _stdio_putc(&_stdout, c);
	if (status < 0) return EOF;
	return (_stdio_flush(&_stdout) > 0) ? 0 : EOF;
}


size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	/* zduvodu bufferovani v _stdio_file se data prenasi po bytech */
        /* TODO: udelat lip */
	
	const unsigned char *data = ptr;
	for (size_t i = 0; i < size * nmemb; i++) {
		int status = _stdio_putc(stream, data[i]);
		if (status < 0) return i / size;
	}

	_stdio_flush(stream);
	return nmemb;
}



/*
 *
 * Prace se vstupy
 *
 */
static inline int _stdio_loadbuf(struct _stdio_file *f)
{
	int size;
	size = sys_read(f->fd, f->inbuf, _STDIO_INBUF_SIZE);
	if (size == 0) f->eof = 1;

	f->inbuf_size = size;
	f->inbuf_idx = 0;

	return size;
}

int getc(FILE *f)
{
	if (f->inbuf_idx >= f->inbuf_size)
		_stdio_loadbuf(f);
	if (f->eof) return EOF;

	return f->inbuf[f->inbuf_idx ++];
}

int getchar()
{
	return getc(&_stdin);
}

char *fgets(char *s, int size, FILE *stream)
{
	int i;
	for (i = 0; i < size - 1; i++) {
		int c = getc(stream);
		if (c == EOF) break;
		s[i] = c;
		if (c == '\n') {
			i++;
			break;
		}
	}
	s[i] = '\0';
	return (i == 0) ? (char *) 0 : s; 
}

size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	/* zduvodu bufferovani v _stdio_file se data prenasi po bytech */
        /* TODO: udelat lip */

	unsigned char *data = ptr;
	for (size_t i = 0; i < size * nmemb; i++) {
		int c = getc(stream);
		if (c == EOF) {
			return i / size;
		}
		*(data++) = c;
	}
	return nmemb;
}

/* 
 *
 * Funkce unixoveho operacniho systemu
 *
 */

int execv(const char *path, char *const argv[])
{
	char *const envp_empty[] = { NULL };
	return sys_execve(path, argv, envp_empty);
}

int sched_yield()
{
	return sys_sched_yield();
}

static void *_cur_brk = 0;

void *sbrk(long increment)
{
	if (!_cur_brk) _cur_brk = sys_brk(0);

	void *old_brk = _cur_brk;
	
	void *expected_brk = ((unsigned char *) _cur_brk) + increment;
	_cur_brk = sys_brk(expected_brk);
	if (_cur_brk != expected_brk) return (void *) -1;
	return old_brk;
}


/*
 *
 * Prace s pameti
 * TODO: udelat slucovani sousednich bloku
 * TODO: pouzit vhodnejsi strukturu pro hledani vetsich bloku 
 *
 */

/* vlastnosti alokatoru -- platformove zavisle */
#define SMALL_BLOCKS_CNT  (32)
#define SMALL_BLOCKS_STEP (8)
#define ROUND_UP(x)	  ((x + 7) & ~7)
#define MIN_ALLOCATION	  (4096 * 16)
#define HDR_SIZE	  (8)

#define BLK_HDR_SIZE(x) 	(*((unsigned long *) ((void **)x) - 1))
#define BLK_HDR_PREV_SIZE(x) 	(*((unsigned long *) ((void **)x) - 2))
#define BLK_SIZE(x) 	        (BLK_HDR_SIZE(x) & ~7)
#define BLK_PREV(x) 	        (*(((void **)x) + 1))
#define BLK_NEXT(x) 	        (*(((void **)x) + 2))


/* seznam volnych bloku */
static void *free_blocks[SMALL_BLOCKS_CNT];

/* jeden spolecny seznam velkych bloku */
static void *large_blocks = NULL;

static size_t prev_alloc_size       = 0;
static unsigned char *data_area     = NULL;
static unsigned long data_area_size = 0;

/** vybere odpovidaji frontu pro danou velikost objektu */
static inline void **__select_queue(size_t size)
{
	int queue_idx = size / SMALL_BLOCKS_STEP;
	if (queue_idx < SMALL_BLOCKS_CNT) {
		return &free_blocks[queue_idx];
	} else {
		return &large_blocks;
	}
}

/** ulozi blok do dane fronty volnych bloku*/
static inline void __release_block(void **queue, void *ptr)
{
	if (*queue == NULL) {
		*queue = ptr;
		BLK_PREV(ptr) = ptr;
		BLK_NEXT(ptr) = ptr;
	} else {
		void *q = *queue;

		BLK_PREV(ptr) = BLK_PREV(q);
		BLK_NEXT(BLK_PREV(q)) = ptr;
		BLK_NEXT(ptr) = q;
		BLK_PREV(q) = ptr;

	}
	*queue = ptr;
}

/** odebere volny blok z fronty volnych bloku */
static inline void __reclaim_block(void **queue, void *free_block)
{
	if (BLK_NEXT(free_block) == free_block) {
		*queue = NULL;
	} else {
		if (free_block == *queue) 
			*queue = BLK_NEXT(free_block);
		BLK_NEXT(BLK_PREV(free_block)) = BLK_NEXT(free_block);
		BLK_PREV(BLK_NEXT(free_block)) = BLK_PREV(free_block);
	}
}

void *malloc(size_t size)
{
	size = ROUND_UP(size);

	/* cast hledajici vhodny jiz uvolneny blok */
	void *free_block = NULL;
	void **queue = __select_queue(size);
	if (*queue) {
		if (queue != large_blocks) {
			free_block = *queue;
		} else {
			/* vyhledava strategii typu first fit */
			void *i = large_blocks;
			do {
				if (BLK_SIZE(i) >= size)
					free_block = i;
				i = BLK_NEXT(i);
			} while (!free_block && (i != large_blocks));

		}
	}
	if (free_block) {
		__reclaim_block(queue, free_block);
		return free_block;
	}

	/* alokuje blok pameti primo */
	unsigned long full_size = size + HDR_SIZE;

	if (data_area_size <= full_size) {
		/* pokud neni volne misto, vyzada si jej od OS */
		unsigned long request_size = 0;
		while ((data_area_size + request_size) < full_size)
			request_size += MIN_ALLOCATION;

		void *start_addr = sbrk(request_size);
		if (!data_area) data_area = start_addr;
		data_area_size += request_size;
	}

	free_block = data_area + HDR_SIZE;
	BLK_HDR_SIZE(free_block) = size;

	data_area       += full_size;
	data_area_size  -= full_size;
	prev_alloc_size  = size;

	return free_block;
}

void free(void *ptr)
{
	void **queue = __select_queue(BLK_HDR_SIZE(ptr));
	__release_block(queue, ptr);
}
