#include <stdint.h>

#include "pg_alloc.h"

#include "../util/list.h"

#define PAGE_SIZE	(4096)
#define PG_INDEX(p)	((p)->index)

static int size_to_order(uint32_t size);

/**
 * Vrati volnou stranku daneho radu, pokud neexistuje, vraci NULL.
 */
static struct page *free_list_get_page(struct pg_zone *zone, int order)
{
	if (order > MAX_ORDER) return NULL;
	struct free_page *result = zone->free_list[order];
	if (!result) return NULL;

	LIST_REMOVE(zone->free_list[order], result);
	return result->page;
}

/**
 * Do seznamu volnych stranek vrati danou stranku,
 * neprovadi slucovani.
 */
static void page_put(struct pg_zone *zone, struct page *page)
{
	struct free_page *fp = (struct free_page *) page->addr;
	fp->page = page;
	page->free = 1;

	LIST_INSERT(zone->free_list[page->order], fp);
}


/**
 * Rozdeli stranku daneho radu,
 * pokud takova stranka neexistuje, rozdeli stranku vyssiho radu.
 * Vraci 0/1 podle toho, jestli se stranku podarilo rozdelit, nebo ne.
 */
static int page_split(struct pg_zone *zone, int order)
{
	if (order > MAX_ORDER) return 0;
	struct page *page_to_split;
	do {
		page_to_split = free_list_get_page(zone, order);
		if (!page_to_split) {
			if (page_split(zone, order + 1)) continue;
			else return 0;
		}
		
		int new_order = --page_to_split->order;

		struct page *second_part = page_to_split + (1 << new_order);
		second_part->order = new_order;

		page_put(zone, page_to_split);
		page_put(zone, second_part);
	} while (!page_to_split);
	return 1;
}

/**
 * Alokuje stranku dane velikosti, size musi byt mocninou 2.
 */
struct page *pg_alloc(struct pg_zone *zone, uint32_t size)
{
	int order = size_to_order(size);
	struct page *result;

	do {
		result = free_list_get_page(zone, order);
		if (!result) {
			if (!page_split(zone, order + 1)) 
				return NULL;
		}
	} while (!result);
	result->free = 0;
	return result;
}

/**
 * Uvolni danou stranku a spoji ji s jejim buddym,
 * pokud existuje. 
 */
void pg_free(struct pg_zone *zone, struct page *page)
{
	/* koncova podminka rekurze */
	if (page->order == MAX_ORDER) {
		page_put(zone, page);
		return;
	}

	/* identifikuje sousedni stranku */
	int pg_idx_1 = PG_INDEX(page);
	int pg_idx_2 = pg_idx_1 ^ (1 << page->order);
	
	struct page *buddy = &zone->pages[pg_idx_2];
	if (buddy->free) {
		/* pokud je sousedni stranka volna, slouci je */
		struct free_page *buddy_fp = (struct free_page *) buddy->addr;
		LIST_REMOVE(zone->free_list[buddy->order], buddy_fp);
		
		int min_idx = pg_idx_1 < pg_idx_2 ? pg_idx_1 : pg_idx_2;
		struct page *coalesced_page = &zone->pages[min_idx];
		coalesced_page->order++;

		/* rekurzivne uvolnuje sloucene stranky */
		pg_free(zone, coalesced_page);
	} else {
		/* vlozi stranku do seznamu volnych stranek */
		page_put(zone, page);
	}
}

/**
 * Uvolni danou pamet. 
 */
void pg_free_mem(struct pg_zone *zone, void *mem)
{
	unsigned long p_idx = ((unsigned long) mem - (unsigned long) zone->base) / PAGE_SIZE;
	pg_free(zone, &(zone->pages[p_idx]));
}

/**
 * Prevede velikost v bytech na order,
 * tj. urci n takove, ze 2^n = size
 */
static int size_to_order(uint32_t size)
{
	size /= PAGE_SIZE;
	for (int i = 31; i >= 0; i--) {
		if (size & (1 << i)) return i;
	}
	return 0;
}


/**
 * Inicializuje zonu se strankami.
 * page_count -- pocet stranek (musi byt mocnina dvou)
 * page -- pole struktur popisujici jednotlive stranky
 * base -- adresa, kde jsou stranky ulozeny
 */
void pg_zone_init(struct pg_zone *zone, int page_count, struct page *pages, unsigned char *base)
{
	zone->base      = base;
	zone->pages     = pages;
	zone->max_order = size_to_order(page_count * PAGE_SIZE);

	/* inicializuje vlasnostni stranek */
	for (int i = 0; i < page_count; i++) {
		pages[i].addr = base + (i * PAGE_SIZE);
		pages[i].order = 0;
		pages[i].index = i;
	}

	/* vynuluje seznam volnych bloku */
	for (int i = 0; i < MAX_ORDER; i++) {
		zone->free_list[i] = NULL;
	}

	/* vlozi nejvetsi stranku */
	pages[0].order = zone->max_order;
	page_put(zone, &pages[0]);
}
