/*
 *
 * Sprava strankovane pameti
 * 
 */

#include <stdint.h>

#include "../util/util.h"
#include "../tasks/tasks.h"

#include "pg_alloc.h"
#include "vm.h"

/* zarovna hodnotu x na nasobek mocniny dvou */
#define ALIGN(x, n)       (((x) + (n) - 1) & (~((n) - 1)))

/* zarovna hodnotu na velikost stranky */
#define ALIGN_PG_SIZE(x)  ALIGN(x, PAGE_SIZE)

#define PG_PRESENT     (0x01)             /* stranka je aktivni */
#define PG_RW          (0x02)             /* do stranky je mozne zapisovat i cst */
#define PG_USER        (0x04)             /* stranka je pristupna v nepriviligenovanem rezimu */

extern struct pg_zone kernel_mem_zone;    /* halda jadra */
uint32_t *vm_active_pd;                   /* aktivni adresar tabulek stranek */

/**
 * Zasobnik volnych ramcu
 */
struct {
	uint32_t *frames;                 /* pole volnych ramcu */
	unsigned int cnt;                 /* pocet ramcu celkem */ 
	unsigned int idx;                 /* ukazatel na dalsi volny ramec */
} free_frames;

static void allocate_free_frames(uint32_t pmem_size);
static inline void invlpg(void* m);
static void out_of_memory_error();
static uint32_t allocate_frame();
static int vm_reclaim_pt(uint32_t pdi, uint32_t *pt);

/**
 * Inicializuje virtualni pamet.
 * Argument udava velikost fyzicke pameti.
 */
void vm_init(uint32_t pmem_size)
{
	allocate_free_frames(pmem_size);
}

/**
 * Vytvori adresar defaultni tabulku stranek,
 * ktera obsahuje pouze stranky nalezici jadru.
 */
uint32_t *vm_create_initial_page_directory() 
{
	/* alokuje adresar tabulek stranek a tabulku stranek pro jadro */
        struct page *initial_table_pg = pg_alloc(&kernel_mem_zone, PAGE_SIZE * 2);
        uint32_t *pd = (uint32_t *) initial_table_pg->addr;
        uint32_t *pt = (uint32_t *) (initial_table_pg->addr + PAGE_SIZE);

        for(int i = 0; i < 1024 * 2; i++)
                pd[i] = 0;

        pd[0] = (((uint32_t) pt) & 0xfffff000) | PG_PRESENT | PG_RW | PG_USER;
        for (int i = 0; i < 1024; i++) {
                if (i * PAGE_SIZE < KERNEL_MEM_END) {
                        pt[i] = (i * PAGE_SIZE) | PG_PRESENT | PG_RW;
                } else {
                        pt[i] = 0;
                }
        }

	return pd;
}

/**
 * Funkce je vyvolana v pripade, ze dojde k pristupu na neplatnou stranku 
 */
void vm_pagefault_handler(uint32_t vaddr, uint32_t flags)
{
	if (vaddr < KERNEL_MEM_END) {
		kprint("Invalid access to address: ");
		kprint_x(vaddr);
		kprint("\r\nProcess terminated.\r\n");
		task_terminate();
	}

	uint32_t *pd = vm_active_pd;

	int pdi = (vaddr >> 22) & 0x3ff;
	int pti = (vaddr >> 12) & 0x3ff;

	uint32_t *pt = (uint32_t *) (pd[pdi] & 0xfffff000);

	/* pokud neexistuje zaznam v adresari tabulek stranek, je vytvoren vcetne nove tabulky stranek */
	if (pd[pdi] == 0) {
		/* tabulka stranek se alokuje v pameti jadra */
		struct page *pt_page = pg_alloc(&kernel_mem_zone, PAGE_SIZE);
		if (!pt_page) {
			out_of_memory_error();
		}

		/* nastavi tabulku stranek */
		pt = (uint32_t *) pt_page->addr;
		pd[pdi] = ((uint32_t) pt) | PG_PRESENT | PG_RW | PG_USER;;

		/* vynuluje obsah tabulky stranek */
		for (int i = 0; i < 1024; i++) 
			pt[i] = 0;
	}
	
	/* ziska volny ramec a vynuluje ho*/
	uint32_t paddr = allocate_frame();
	if (paddr == 0) { 
		out_of_memory_error();
	}

	/* nastavi zaznam v tabulce stranek a vymaze zaznam v TLB */
	pt[pti] = (paddr & 0xfffff000) | PG_PRESENT | PG_RW | PG_USER;
	invlpg((void *) vaddr);

	/* vymaze obsah stranky */
	uint32_t *page = (uint32_t *) (pdi << 22 | pti << 12);
	for (int i = 0; i < 1024; i++)
		page[i] = 0;
	

/*      -- ukaze, jak jsou mapovany logicke adresy na fyzicke
	kprint_x(vaddr);
	kprint(" -> ");
	kprint_x(paddr);
	kprint("\r\n");
*/
}

/**
 * Uvolni cely adresar tabulek stranek.
 */
void vm_reclaim(uint32_t *pd)
{
	for (int i = 0; i < 1024; i++) {
		if (pd[i]) {
			uint32_t pt_addr = pd[i] & 0xfffff000; 
			int empty = vm_reclaim_pt(i, (uint32_t *) pt_addr);
			if (empty) {
				pg_free_mem(&kernel_mem_zone, (void *) pt_addr);
				pd[i] = 0;
			}
		}
	}
}

/**
 * Alokuje zasobnik volnych ramcu.
 */
static void allocate_free_frames(uint32_t pmem_size)
{
	/* pocet ramcu, ktere budou k dispozici */
	free_frames.cnt = pmem_size / PAGE_SIZE;

	/* v pameti jadra si alokuje souvisly blok, ktery bude slouzit jako pole */
	int free_frames_pages_size = ALIGN_PG_SIZE(free_frames.cnt * sizeof(uint32_t));
	struct page *free_frames_pages = pg_alloc(&kernel_mem_zone, free_frames_pages_size); 
	if (!free_frames_pages) {
		kprint("Out of memory\r\n");
		while (1);
	}
	free_frames.frames = (uint32_t *) free_frames_pages->addr;

	/* do tohoto pole ulozi adresy volnych ramcu */
	for (int i = 0; i < free_frames.cnt; i++) {
		free_frames.frames[i] = KERNEL_MEM_END + i * PAGE_SIZE;
	}
	free_frames.idx = 0;

}

/**
 * Vraci fyzickou adresu volneho ramce.
 */
static uint32_t allocate_frame()
{
	if (free_frames.idx == free_frames.cnt) return 0;
	return free_frames.frames[free_frames.idx++];
}

/** 
 * Ramec dany fyzickou adresou oznaci jako znovu pouzitelny.
 */
static void release_frame(uint32_t paddr)
{
	free_frames.frames[--free_frames.idx] = paddr;
}


/**
 * Vypise chybovou hlasku a ukonci proces
 */
static void out_of_memory_error()
{
	kprint("Out of memory. Process terminated.\r\n");
	task_terminate();
}

/**
 * Uvolni strankovaci tabulku, vraci true, pokud je tabulka zcela prazdna.
 */
static int vm_reclaim_pt(uint32_t pdi, uint32_t *pt)
{
	int act_entries = 0;
	for (int i = 0; i < 1024; i++) {
		if (pt[i]) {
			uint32_t paddr = pt[i] & 0xfffff000;
			uint32_t vaddr = pdi << 22 | i << 12;
			/* ramce pratrici jadru ignorujeme */
			if (paddr >= KERNEL_MEM_END) {
				release_frame(paddr);
				pt[i] = 0;
				invlpg((void *) vaddr);
			} else {
				act_entries++;
			}
		}
	}
	return !act_entries;
}

/**
 * Zneplatni zaznam v TLB.
 */
static inline void invlpg(void* m)
{
	__asm__ volatile ("invlpg (%0)" : : "b"(m) : "memory" );
}
