#include "../fs/mufs.h"
#include "../util/string.h"
#include "../util/util.h"
#include "../mem/vm.h"
#include "../irq/irq.h"
#include "../mem/alloc.h"
#include "../mem/pg_alloc.h"
#include "readelf.h"
#include "tasks.h"

#define MAX_TASKS	        (16)             /* maximalni pocet bezicich procesu */

#define KERNEL_STACK_SIZE	(16 * PAGE_SIZE) /* velikost jaderneho zasobniku */
#define USER_STACK              (0x003f0000)     /* zacatek zasobniku uzivatelskych programu */
#define TSS                     (0x80000)        /* adresa TSS */
#define MAX_ARGS_SIZE	        (255)            /* maximalni pocet argumentu */
#define TIMER_TICKS		(60000)          /* pocet tiku casovace */
#define PROCESS_QUANTUM		(20)             /* casove akvantu (mereno v poctu aktivaci casovace), pridelene kazdemu procesu */
#define COMMAND_SHELL	        "/cosh"          /* cesta k programu s prikazovou radkou */

//#define COMMAND_SHELL	        "/hello02"       /* cesta k programu s prikazovou radkou */
//#define COMMAND_SHELL	        "/loop"          /* cesta k programu s prikazovou radkou */

struct task *task_running;                 /* aktualne bezici uloha */
struct task *task_ready[MAX_TASKS];        /* prichystane ulohy */
int task_ready_count;                      /* pocet pripravenych uloh */
int task_ready_head, task_ready_tail;      /* ukazatel do fronty pripravenych uloh */

struct kmem_cache *mem_task;               /* alokuje struktury pro jednotlive procesy */
struct kmem_cache *mem_args;               /* alokuje bloky o velikosti 256 B pro ulozeni argumentu program */
extern struct mufs root_fs;                /* korenovy souborovy system */

extern struct pg_zone kernel_mem_zone;     /* halda jadra */

/* pokud tato promenna odkazuje na adresar tabulek stranek, je pri nasledujici prenastaveni uvolnen */
static uint32_t *free_page_directory;    

static inline void timer_activate();
static void args_store(struct task *task, char *path, char **argv);
static void args_free(struct task *task);
static void args_pass(struct task *task);


/**
 * Inicializuje struktury nutne pro planovani procesu.
 */
void task_init()
{
	mem_task = kmem_cache_create(sizeof(struct task), NULL, NULL, NULL);
	mem_args = kmem_cache_create(MAX_ARGS_SIZE + 1, NULL, NULL, NULL);
	task_running = NULL;
	task_ready_count = 0;
}

/**
 * Vytvori prazdnou strukturu procesu, bez kodu, ktery by se mel provadet.
 */
static struct task *task_create()
{
	// TODO: testovat, jestli nedoslo k prekroceni poctu vytvorenych procesu
	struct task *t    = kmem_cache_alloc(mem_task);
	t->page_directory = vm_create_initial_page_directory();
	t->kernel_stack   = pg_alloc(&kernel_mem_zone, KERNEL_STACK_SIZE)->addr; /* zasobnik roste zhora dolu */
	t->esp            = (void *) ((uint32_t) t->kernel_stack) + KERNEL_STACK_SIZE - 4;
	for (int i = 0; i < MAX_FILES_OPENED; i++) {
		t->fd[i] = NULL;
	}
	return t;
}


/**
 * Aktivuje adresar tabulek stranek.
 */
static void switch_page_directory(uint32_t *pd)
{
	vm_set_page_directory(pd);
	if (free_page_directory) {
		pg_free_mem(&kernel_mem_zone, (void *) free_page_directory);
		free_page_directory = NULL;
	}
}

/**
 * Prenastavi adresu jaderneho zasobniku pro dalsi preruseni.
 */
static void switch_kernel_stack(void *addr)
{
	uint32_t *tss = (uint32_t *) TSS;
	tss[1] = (uint32_t) addr;
}


/**
 * Pro aktulni proces nahraje do pameti kod a data ze souboru daneho cestou path. 
 */
static int task_load(char *path)
{
	Elf32_Ehdr ehdr;
	Elf32_Phdr phdr;

	struct mufs_fd *f = mufs_file_open(&root_fs, path, 0);
	if (!f) return EINVALID_FILE;
	
	int status = readelf_file_header(f, &ehdr);
	if (status == EOK) {
		unsigned long min_brk = 0;
		for (int i = 0; i < ehdr.e_phnum; i++) {
			readelf_program_header(f, i, &ehdr, &phdr);
			mufs_file_seek(f, phdr.p_offset, SEEK_SET);

			if (phdr.p_type != PT_LOAD) continue;

			if (phdr.p_vaddr < KERNEL_MEM_END) {
				kprint("Attempt to corrupt kernel memory\r\n");
				while (1);
			}

			memset((void *) phdr.p_vaddr, 0x00, phdr.p_memsz);
			mufs_file_read(f, (unsigned char *) phdr.p_vaddr, phdr.p_filesz);
			
			if (phdr.p_vaddr + phdr.p_memsz > min_brk) {
				min_brk = phdr.p_vaddr + phdr.p_memsz;
			}
		}
		task_running->cur_brk = min_brk;
		task_running->min_brk = min_brk;
		task_running->entry_point = (void (*)()) ehdr.e_entry;
	} else {
		kprint("Invalid file: ");
		kprint(path);
		kprint("\r\n");
		return EINVALID_FILE;
	}
	mufs_file_close(f);
	return EOK;
}

/**
 * Spusti shell
 */
void task_exec_shell()
{
	char *args[] = { COMMAND_SHELL, NULL };
	task_exec(COMMAND_SHELL, args);
}

/**
 * Zaradi zadany proces do fronty cekajicich procesu.
 */
static inline void task_enqueue(struct task *task)
{
	if (task_ready_count == 0) {
		task_ready_tail = task_ready_head = 0;
	} else {
		task_ready_tail = (task_ready_tail + 1) % MAX_TASKS;
	}
	task_ready[task_ready_tail] = task;
	task_ready_count ++;
}

/**
 * Odebere z fronty jeden proces, pokud takovy neni, vraci NULL.
 */
static inline struct task *task_dequeue()
{
	if (!task_ready_count) return NULL;
	struct task *result  = task_ready[task_ready_head];
	task_ready_head = (task_ready_head + 1) % MAX_TASKS;
	task_ready_count--;
	return result;
}

/**
 * Spusti novy proces se zadanymi argumenty.
 */
void task_exec(char *path, char **argv)
{
	struct task *task = task_create();

	args_store(task, path, argv); 

	task->status = TASK_NEW;
	task_enqueue(task);
}

/**
 * Aktivuje proces, ktery dosud jeste nebezel.
 */
static int run_new_task(struct task *task)
{
	switch_page_directory(task->page_directory);
	switch_kernel_stack(task->esp);

	int status = task_load(task->path);
	if (status != EOK) return EINVALID_FILE;
	
	args_pass(task);
	timer_activate();
	jumpto_task((unsigned long) task->entry_point);

	return EOK;
}

/**
 * Funkce zavolana pri volani casovace (preemptivni preruseni behu procesu
 */
void task_preempt()
{
	task_running->switch_type = ST_PREEMPT;
	task_scheduler();
}

/**
 * Funkce zavolana pri prepnuti procesu, ktery je prave v jadre.
 */
void task_kernel_switch()
{
	task_running->switch_type = ST_KERNEL_YIELD;
	task_scheduler();
}


/**
 * Deaktivuje aktualne bezici ulohu.
 */
static void deactive_running_task()
{
	/* pokud doslo k pokusu o preemptivni prepnuti ulohy, 
         * sleduje se, kolikrat k tomu doslo a pripadne
         * je proces znovu aktivovan */
	if (task_running->switch_type == ST_PREEMPT) {
		task_running->ticks--;
		if (task_running->ticks >= 0) {
			timer_activate();
			task_resume_preempt();
		}
	}

	/* zarazeni ulohy do fronty */
	task_running->status = TASK_READY;
	task_enqueue(task_running);
}

/**
 * Funkce odebere z fronty jednu ulohu a necha ji vykonavat.
 */
static void run_next_task()
{
	task_running = task_dequeue();
	task_running->ticks = PROCESS_QUANTUM;
	if (task_running->status == TASK_NEW) {

		/* je spustena zcela nova uloha */
		task_running->status = TASK_RUNNING;
		int status = run_new_task(task_running);
		if (status != EOK) {
			task_terminate();
		}

	} else {
		/* je spustena jiz jednou aktivni uloha */
		switch_page_directory(task_running->page_directory);
		switch_kernel_stack(task_running->kernel_stack);

		timer_activate();
		switch (task_running->switch_type) {
			case ST_YIELD: task_resume_yield(); break;
			case ST_PREEMPT: task_resume_preempt(); break;
			case ST_KERNEL_YIELD: task_resume_kernel_yield(); break;
		}
	}
	kprint("Fatal error\r\n");
	while (1);
}

/**
 * Funkce planovace uloh.
 */
static int initial_run = 1; /* indikuje, ze planovac jeste nebyl spusten */
void task_scheduler()
{
	/* pokud je planovac aktivovan poprve, spusti shell */
	if (initial_run) {
		kprint("Running initial shell\r\n");
		initial_run = 0;
		task_exec_shell();
	}

	while (1) {
		/* odsune aktualne bezici proces */
		if (task_running != NULL) {
			deactive_running_task();
		}

		/* aktivuje proces, cekajici ve fronte */ 
		if (task_ready_count > 0) {
			run_next_task();
		}
	}
}

/**
 * Zavre vsechny otevrene soubory aktualniho procesu
 */
static void close_all_files()
{
	for (int i = 0; i < MAX_FILES_OPENED; i++) {
		if (task_running->fd[i] != NULL) {
			mufs_file_close(task_running->fd[i]);
			task_running->fd[i] = NULL;
		}
	}
}

/**
 * Ukonci aktualne bezici proces a uvolni vsechny jeho zdroje.
 */
void task_terminate()
{
	close_all_files();
	vm_reclaim(task_running->page_directory);
	free_page_directory = task_running->page_directory; 
	
	args_free(task_running);
	pg_free_mem(&kernel_mem_zone, task_running->kernel_stack);
	kmem_cache_free(mem_task, task_running);

	task_running = NULL;
	task_scheduler();
}

/**
 * Aktivuje casovac
 */
static inline void timer_activate()
{
	pit_set(TIMER_TICKS);
}

/**
 * Funkce ulozi argumenty predavane programu do struktury jadra,
 * ne se zmeni rozlozeni pameti.
 */
static void args_store(struct task *task, char *path, char **argv)
{
	task->path = kmem_cache_alloc(mem_args);
	strlcpy(task->path, path, MAX_ARGS_SIZE);

	int i = 0;
	while (argv[i] && (i < MAX_ARGS)) {
		task->argv[i] = kmem_cache_alloc(mem_args);
		strlcpy(task->argv[i], argv[i], MAX_ARGS_SIZE);
		i++;
	}
	task->argc = i;
}

/**
 * Uvolni argumenty ulozene v jadre
 */
static void args_free(struct task *task)
{
	kmem_cache_free(mem_args, task->path);
	for (int i = 0; i < task->argc; i++) 
		kmem_cache_free(mem_args, task->argv[i]);
}

/**
 * Ulozi argumenty pro spousteny program na zasobnik.
 */
static void args_pass(struct task *task)
{
	unsigned long *stack        = (unsigned long *) USER_STACK;

	/* v teto casti jsou umisteny hodnoty, pred ni nasleduje oblast s odkazy */
	unsigned char *stacked_args = (unsigned char *) (stack + MAX_ARGS);

	for (int i = 0; i < task->argc; i++) {
		char *arg = task->argv[i];
		int len   = strlen(arg);
		stack[i + 1] = (unsigned long) stacked_args;

		/* prekopiruje argument na zasobnik, v jadre neni pristupny z ring3 */
		strcpy((char *) stacked_args, arg);
		stacked_args += len + 1;
	}

	stack[0] = task->argc;
}


